"""Class that handles the parameters of a MaBoSS simulation.
"""

from __future__ import print_function
import collections
from functools import reduce
from sys import stderr, stdout, version_info
if version_info[0] < 3:
    from contextlib2 import ExitStack
else:
    from contextlib import ExitStack

from colomoto import ModelState

from .result import Result
from .cmabossresult import CMaBoSSResult
import os
import uuid
import subprocess
import tempfile
import shutil
from colomoto_jupyter import import_colomoto_tool
from colomoto_jupyter.sessionfiles import new_output_file

_default_parameter_list = collections.OrderedDict([
    ('time_tick', 0.1),
    ('max_time', 4),
    ('sample_count', 10000),
    ('discrete_time', 0),
    ('use_physrandgen', 1),
    ('seed_pseudorandom', 0),
    ('display_traj', 0),
    ('statdist_traj_count', 0),
    ('statdist_cluster_threshold', 1),
    ('thread_count', 1),
    ('statdist_similarity_cache_max_size', 20000)
])

class Simulation(object):
    """
    Class that handles MaBoSS simulations.


    .. py:attribute:: network

      A Network object, that will be translated in a bnd file

    .. py:attribute:: mutations

      A list of nodes for which mutation can be triggered by
      modifying the cfg file

    .. py:attribute:: palette

      A mapping of nodes to color for plotting the results of
      the simulation.

    .. py:attribute:: param

      A dictionary that contains global variables (keys starting with a '$'),
      and simulation parameters (keys not starting with a '$').
    """


    def __init__(self, nt, parameters=collections.OrderedDict({}), command=None, **kwargs):
        """
        Initialize the Simulation object.

        :param nt: the network associated with the simulation.
        :type nt: :py:class:`Network`
        :param dict kwargs: parameters of the simulation
        """
        self.param = _default_parameter_list.copy()
        self.palette = {}
        for cfg in (parameters, kwargs):
            for p in cfg:
                if p in _default_parameter_list or p[0] == '$':
                    self.param[p] = cfg[p]
                elif p == "palette":
                    self.palette = kwargs[p]
                else:
                    print("Warning: unused parameter %s" % p, file=stderr)

        self.network = nt
        self.mutations = []
        self.mutationTypes = {}
        self.refstate = {}

        self.workdir = None
        self.overwrite = False

        errors = self.check(command=command)
        if len(errors) > 0:
            print(errors)

    def update_parameters(self, **kwargs):
        """Add elements to ``self.param``."""
        for p in kwargs:
            if p in _default_parameter_list or p[0] == '$':
                self.param[p] = kwargs[p]
            else:
                print("Warning: unused parameter %s" % p, file=stderr)

    def copy(self):
        new_network = self.network.copy()
        result = Simulation(new_network, self.param, palette=self.palette)
        if self.mutations:
            result.mutations = self.mutations.copy()
        return result

    def check(self, command=None):

        try:
            path = tempfile.mkdtemp()
            cfg_fd, cfg_path = tempfile.mkstemp(dir=path, suffix='.cfg')
            os.close(cfg_fd)
            bnd_fg, bnd_path = tempfile.mkstemp(dir=path, suffix='.bnd')
            os.close(bnd_fg)

            with ExitStack() as stack:
                bnd_file = stack.enter_context(open(bnd_path, 'w'))
                cfg_file = stack.enter_context(open(cfg_path, 'w'))
                self.print_bnd(out=bnd_file)
                self.print_cfg(out=cfg_file)

            maboss_cmd = self.get_maboss_cmd()
            if command is not None:
                maboss_cmd = command

            proc = subprocess.Popen(
                [maboss_cmd, "--check", "-c", cfg_path, bnd_path],
                cwd=path, stdout=subprocess.PIPE, stderr=subprocess.PIPE
            )
            (t_stdout, t_stderr) = proc.communicate()

            shutil.rmtree(path)

            messages = []
            if len(t_stderr) != 0:
                messages = [mess.replace("MaBoSS: ", "") for mess in list(set(t_stderr.decode().split("\n"))) if mess != '']

            return messages

        except ValueError as e:
            return [str(e)]

    def get_maboss_cmd(self):

        maboss_cmd = "MaBoSS"

        l = len(self.network)
        assert l <= 1024, "Models with more than 1024 nodes are not compatible with this version of MaBoSS"

        if l <= 64:
            pass
        elif l <= 128:
            maboss_cmd = "MaBoSS_128n"
        elif l <= 256:
            maboss_cmd = "MaBoSS_256n"
        elif l <= 512:
            maboss_cmd = "MaBoSS_512n"
        else:
            maboss_cmd = "MaBoSS_1024n"

        return maboss_cmd

    def get_cmaboss(self):
        l = len(self.network)
        assert l <= 1024, "Models with more than 1024 nodes are not compatible with this version of MaBoSS"

        if l <= 64:
            return __import__("cmaboss")
        elif l <= 128:
            return __import__("cmaboss_128n")
        elif l <= 256:
            return __import__("cmaboss_256n")
        elif l <= 512:
            return __import__("cmaboss_512n")
        else:
            return __import__("cmaboss_1024n")

    def print_bnd(self, out=stdout):
        """Produce the content of the bnd file associated to the simulation."""
        print(self.network, file=out)

    def print_cfg(self, out=stdout):
        """Produce the content of the cfg file associated to the simulation."""
        print(self.str_cfg(), file=out)

    def str_cfg(self):

        res = "$nb_mutable = %d;\n" % len(self.mutations)
        for p in self.param:
            if p[0] == '$':
                res += "%s = %s;\n" % (p, self.param[p])

        res += self.network.str_istate() + "\n"
        res += "\n"

        for p in self.param:
            if p[0] != '$':
                res += "%s = %s;\n" % (p, self.param[p])

        for name in self.network.names:
            res += "%s.is_internal = %s;\n" % (name, self.network[name].is_internal)

        for nd in self.refstate:
            res += "%s.refstate = %s;\n" % (nd, self.refstate[nd])

        return res

    def get_logical_rules(self):

        path = tempfile.mkdtemp()
        cfg_fd, cfg_path = tempfile.mkstemp(dir=path, suffix='.cfg')
        os.close(cfg_fd)
        bnd_fd, bnd_path = tempfile.mkstemp(dir=path, suffix='.bnd')
        os.close(bnd_fd)
        
        with ExitStack() as stack:
            bnd_file = stack.enter_context(open(bnd_path, 'w'))
            cfg_file = stack.enter_context(open(cfg_path, 'w'))
            self.print_bnd(out=bnd_file)
            self.print_cfg(out=cfg_file)

        proc = subprocess.Popen(
            [self.get_maboss_cmd(), "-c", cfg_path, "-l", bnd_path],
            stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        out, err = proc.communicate()

        if proc.returncode != 0:
            raise Exception(err.decode())
            
        rules = {}
        for line in out.decode().split("\n"):
            if ":" in line:
                node, rule = line.split(" : ", 1)
                rules.update({node.strip(): rule.strip()})

        return rules



    def run(self, command=None, workdir=None, overwrite=False, prefix="res", cmaboss=False, only_final_state=False):
        """Run the simulation with MaBoSS and return a Result object.

        :param command: specify a MaBoSS command, default to None for automatic selection
        :rtype: :py:class:`Result`
        """
        if cmaboss:
            return CMaBoSSResult(self, workdir=workdir, overwrite=overwrite, only_final_state=only_final_state)

        if workdir is not None:
            self.workdir = workdir

        if overwrite:
            self.overwrite = overwrite
            
        return Result(self, command, self.workdir, self.overwrite, prefix)


    def mutate(self, node, state):
        """
        Trigger or untrigger mutation for a node.

        :param node: The :py:class:`Node` to be modified
        :type node: :py:class:`Node`
        :param str State:

            * ``'ON'`` (always up)
            * ``'OFF'`` (always down)
            * ``'WT'`` (mutable but with normal behaviour)


        The node will appear as a mutable node in the bnd file.
        This means that its rate will be of the form:

        ``rate_up = $LowNode ? 0 :($HighNode ? 1: (@logic ? rt_up : 0))``

        If the node is already mutable, this method will simply set $HighNode
        and $LowNode accordingly to the desired mutation.
        """
        if node not in self.network:
            print("Error, unknown node %s" % node, file=stderr)
            return

        nd = self.network[node]
        if not nd.is_mutant:
            self.network[node]=_make_mutant_node(nd)
            self.mutations.append(nd.name)
            self.mutationTypes.update({nd.name: state})


        lowvar = "$Low_"+node
        highvar = "$High_"+node
        if state == "ON":
            self.param[lowvar] = 0
            self.param[highvar] = 1


        elif state == "OFF":
            self.param[lowvar] = 1
            self.param[highvar] = 0

        elif state == "WT":
            self.param[lowvar] = 0
            self.param[highvar] = 0

        else:
            print("Error, state must be ON, OFF or WT", file=stderr)
            return

    def continue_from_result(self, result):
        """Set the initial state from as the last state from result."""
        node_prob = result.get_nodes_probtraj()
        nodes = node_prob.iloc[-1]
        for i in nodes.index:
            if i != "<nil>":
                prob = nodes[i]
                self.network.set_istate(i, [1 - prob, prob])

    def get_initial_state(self):
        """
        TODO
        """
        istate = ModelState()
        def _simplify(states):
            if len(states) == 1:
                return states.pop()
            return list(sorted(states))
        for key, val in self.network.get_istate().items():
            if str in map(type, val.values()):
                raise NotImplementedError("This MaBoSS initial state contains macros which is not supported yet")
            states = [i for i, prob in val.items() if prob > 0]
            if isinstance(key, tuple):
                pre = {i: set() for i in key}
                for subs in states:
                    for (i,state) in zip(key, subs):
                        pre[i].add(state)
                count = reduce(int.__mul__, map(len, pre.values()))
                if count != len(states):
                    raise ValueError("This MaBoSS initial state cannot be converted to a single hypercube")
                for key, states in pre.items():
                    istate[key] = _simplify(states)
            else:
                istate[key] = _simplify(states)
        return istate

    def get_mutations(self):
        return self.mutationTypes

    def set_workdir(self, path, overwrite=False):
        self.workdir = path
        self.overwrite = overwrite

def _make_mutant_node(nd):
    """Create a new logic for mutation that can be activated from .cfg file."""
    curent_rt_up = nd.rt_up
    curent_rt_down = nd.rt_down

    lowvar = "$Low_"+nd.name
    highvar = "$High_"+nd.name
    rt_up = (lowvar+" ? 0 : (" + highvar + " ? 1E308/$nb_mutable : ("
             + curent_rt_up + "))")
    rt_down = (highvar + " ? 0 : (" + lowvar + " ? 1E308/$nb_mutable : ("
             + curent_rt_down + "))")
    # Once this is done, the mutation can be activated by modifying the value
    # of $Low_nodename and $High_nodename in the .cfg file
    newNode = nd.copy()
    newNode.rt_up = rt_up
    newNode.rt_down = rt_down
    newNode.is_mutant = True
    return newNode

def set_nodes_istate(masim, nodes, istate):
    for n in nodes:
        masim.network.set_istate(n, istate)

def set_output(masim, output):
    masim.network.set_output(output)

def copy_and_mutate(masim, nodes, mut):
    masim2 = masim.copy()
    for node in nodes:
        masim2.mutate(node, mut)
    return masim2

def copy_and_update_parameters(sim, parameters):
    new_sim = sim.copy()
    new_sim.update_parameters( **parameters )
    return new_sim

def to_biolqm(maboss_model):
    
    biolqm = import_colomoto_tool("biolqm")
    bnet_filename = new_output_file("bnet")
    with open(bnet_filename, "w") as bnet_file:
        for node, rule in maboss_model.get_logical_rules().items():
            bnet_file.write("%s, %s\n" % (node, rule))

    return biolqm.load(bnet_filename)

def to_minibn(maboss_model):
    from colomoto import minibn
    bnet_filename = new_output_file("bnet")
    with open(bnet_filename, "w") as bnet_file:
        for node, rule in maboss_model.get_logical_rules().items():
            bnet_file.write("%s, %s\n" % (node, rule))

    return minibn.BooleanNetwork.load(bnet_filename)
 
def sbml_to_bnd_and_cfg(sbml_filename, bnd_filename, cfg_filename, use_sbml_names=False):
    
    from .gsparser import loadSBML
    model = loadSBML(sbml_filename, use_sbml_names=use_sbml_names, cmaboss=True)
    with open(bnd_filename, "w") as bnd_file:
        model.print_bnd(bnd_file)
    with open(cfg_filename, "w") as cfg_file:
        model.print_cfg(cfg_file)

def bnet_to_bnd_and_cfg(bnet_filename, bnd_filename, cfg_filename):
    
    from .gsparser import loadBNet
    model = loadBNet(bnet_filename, cmaboss=True)
    with open(bnd_filename, "w") as bnd_file:
        model.print_bnd(bnd_file)
    with open(cfg_filename, "w") as cfg_file:
        model.print_cfg(cfg_file)

def sbml_to_maboss(sbml_filename, cfg_filename=None, use_sbml_names=False):
    from .gsparser import load
    bnd_filename = new_output_file("bnd")
    default_cfg_filename = new_output_file("cfg")
    sbml_to_bnd_and_cfg(sbml_filename, bnd_filename, default_cfg_filename, use_sbml_names)
    sim = load(bnd_filename, cfg_filename if cfg_filename is not None else default_cfg_filename)
    if cfg_filename is None:
        for node in sim.network:
            sim.network.set_istate(node, [1, 0])
    return sim
    
def bnet_to_maboss(sbml_filename, cfg_filename=None):
    from .gsparser import load
    bnd_filename = new_output_file("bnd")
    default_cfg_filename = new_output_file("cfg")
    bnet_to_bnd_and_cfg(sbml_filename, bnd_filename, default_cfg_filename)
    sim = load(bnd_filename, cfg_filename if cfg_filename is not None else default_cfg_filename)
    if cfg_filename is None:
        for node in sim.network:
            sim.network.set_istate(node, [1, 0])
    return sim
    


"""
Class that contains the cMaBoSS simulation.
"""
from .sbmlcmabossresult import SBMLCMaBoSSResult
from sys import stdout
class SBMLSSimulation(object):

    def __init__(self, sbml, cfgs=None, use_sbml_names=False):

        self.sbml = sbml
        self.cfgs = cfgs

        self.nb_nodes = self.count_nodes()
        self.cmaboss = self.get_cmaboss()

        if self.cfgs is None:
            self.cmaboss_sim = self.cmaboss.MaBoSSSim(self.sbml, use_sbml_names=use_sbml_names)
        else:
            self.cmaboss_sim = self.cmaboss.MaBoSSSim(self.sbml, self.cfgs, use_sbml_names=use_sbml_names)

        self.network = self.get_logical_rules()
        
    def count_nodes(self):
        
        res = 0
        with open(self.sbml, 'r') as f:
            lines = f.readlines()
            for line in lines:
                if len(line) > 0 and "</qual:qualitativespecies>" in line.lower():
                    res += 1
        return res

    def print_bnd(self, out=stdout):
        """Produce the content of the bnd file associated to the simulation."""
        print(self.cmaboss_sim.str_bnd(), file=out)

    def print_cfg(self, out=stdout):
        """Produce the content of the cfg file associated to the simulation."""
        print(self.cmaboss_sim.str_cfg(), file=out)

    def get_cmaboss(self):

        assert self.nb_nodes <= 1024, "Models with more than 1024 nodes are not compatible with this version of MaBoSS"

        if self.nb_nodes <= 64:
            return __import__("cmaboss")
        elif self.nb_nodes <= 128:
            return __import__("cmaboss_128n")
        elif self.nb_nodes <= 256:
            return __import__("cmaboss_256n")
        elif self.nb_nodes <= 512:
            return __import__("cmaboss_512n")
        else:
            return __import__("cmaboss_1024n")

    def get_logical_rules(self):

        out = self.cmaboss_sim.get_logical_rules()    
        rules = {}
        for line in out.split("\n"):
            if ":" in line:
                node, rule = line.split(" : ", 1)
                rules.update({node.strip(): rule.strip()})

        return rules


    def run(self, only_final_state=False):
        return SBMLCMaBoSSResult(self, only_final_state)


__all__ = ["SBMLSSimulation"]



"""
Class that contains the results of a MaBoSS simulation.
"""

import pandas
import tempfile
import os
from .results.baseresult import BaseResult
from contextlib import ExitStack
from time import time
class SBMLCMaBoSSResult(BaseResult):

    def __init__(self, sim, only_final_state=False):

        BaseResult.__init__(self, "some_path")
        # self.cmaboss_simulation = sim.cmaboss.MaBoSSSim(network=sim.sbml, config=sim.cfgs)
        self.cmaboss_result = sim.cmaboss_sim.run(only_last_state=only_final_state)
        self.simul = sim
        # self.workdir = workdir
        self.only_final_state = only_final_state

        # if workdir is not None:
        #     if not os.path.exists(workdir):
        #         os.makedirs(workdir)

        #     self.cmaboss_result.display_run(os.path.join(workdir, "%s_run.txt" % prefix))

        #     if self.only_final_state:
        #         self.cmaboss_result.display_final_states(os.path.join(workdir, "%s_finalprob.csv" % prefix))
        #     else:
        #         self.cmaboss_result.display_probtraj(os.path.join(workdir, "%s_probtraj.csv" % prefix))
        #         self.cmaboss_result.display_fp(os.path.join(workdir, "%s_fp.csv" % prefix))
        #         self.cmaboss_result.display_statdist(os.path.join(workdir, "%s_statdist.csv" % prefix))


    def get_last_states_probtraj(self):
        
        raw_res, states, timepoints = self.cmaboss_result.get_last_probtraj()
        df = pandas.DataFrame(raw_res, columns=states, index=timepoints)
        return df

    def get_states_probtraj(self, prob_cutoff=None):
        if not self.only_final_state:

            raw_res, states, timepoints = self.cmaboss_result.get_probtraj()
            df = pandas.DataFrame(raw_res, columns=states, index=timepoints)
            df.sort_index(axis=1, inplace=True)

            if prob_cutoff is not None:
                maxs = df.max(axis=0)
                return df[maxs[maxs>prob_cutoff].index]

            return df

    def get_nodes_probtraj(self, prob_cutoff=None):
        if not self.only_final_state:
            raw_res, nodes, timepoints = self.cmaboss_result.get_nodes_probtraj()
            df = pandas.DataFrame(raw_res, columns=nodes, index=timepoints)
            df.sort_index(axis=1, inplace=True)

            if prob_cutoff is not None:
                maxs = df.max(axis=0)
                return df[maxs[maxs>prob_cutoff].index]

            return df

    def get_last_nodes_probtraj(self):
        raw_res, nodes, timepoints = self.cmaboss_result.get_last_nodes_probtraj()
        df = pandas.DataFrame(raw_res, columns=nodes, index=timepoints)
        df.sort_index(axis=0, inplace=True)

        return df


    def get_fptable(self):
        if not self.only_final_state:
            raw_res = self.cmaboss_result.get_fp_table()

            df = pandas.DataFrame(["#%d" % fp for fp in sorted(raw_res.keys())], columns=["FP"])

            df["Proba"] = [raw_res[fp][0] for fp in sorted(raw_res.keys())]
            df["State"] = [raw_res[fp][1] for fp in sorted(raw_res.keys())]

            for node in self.simul.network.keys():
                df[node] = [1 if node in raw_res[fp][1].split(" -- ") else 0 for fp in sorted(raw_res.keys())]

            return df


__all__ = ["SBMLCMaBoSSResult"]



"""Tools to draw figures from the MaBoSS results."""

import operator

import pandas as pd
import matplotlib.pylab as plt
import numpy as np

def persistent_color(palette, state):

    reordered = " -- ".join(sorted(state.split(" -- ")))
    if state in palette:
        return palette[state]
    if reordered in palette:
        return palette[reordered]
    if state == "Others":
        color = "lightgray"
    else:
        count = len(palette)
        if "Others" not in palette:
            count += 1
        color = "C%d" % (count % 10)
    palette[state] = color
    palette[reordered] = color
    return color

def register_states_for_color(palette, collection):
    [persistent_color(palette, state) for state in collection]

def make_plot_trajectory(time_table, ax, palette, legend=True, error_table=None):
    register_states_for_color(palette, time_table.columns.values)
    color_list = [persistent_color(palette, idx) \
                    for idx in time_table.columns.values]
    time_table.plot(ax=ax, color=color_list, legend=legend, yerr=error_table)
    if legend:
        plt.legend(loc=(1.1, 0))


def plot_node_prob(time_table, ax, palette, legend=True, error_table=None):
    """Plot the probability of each node being up over time."""
    register_states_for_color(palette, time_table.columns.values)
    color_list = [persistent_color(palette, idx) \
                    for idx in time_table.columns.values]
    time_table.plot(ax=ax, color=color_list, legend=legend, yerr=error_table)
    if legend:
        plt.legend(loc='upper right')


def plot_piechart(plot_table, ax, palette, embed_labels=False, autopct=4, \
                    prob_cutoff=0.01, legend=True, nil_label=None):
    plot_line = plot_table.iloc[-1].rename("")  # Takes the last time point

    others = plot_line[plot_line <= prob_cutoff].sum()

    plot_line = plot_line[plot_line > prob_cutoff]
    plot_line.sort_values(ascending=False, inplace=True)

    if others:
        plot_line.at["Others"] = others

    register_states_for_color(palette, plot_line.index)

    plotting_labels = []
    color_list = []
    for value_index, value in enumerate(plot_line):
        state = plot_line.index.values[value_index]
        color_list.append(persistent_color(palette, state))
        if embed_labels and value >= 0.1:
            if state == "<nil>" and nil_label is not None:
                plotting_labels.append(nil_label)
            else:
                plotting_labels.append(state)
        else:
            plotting_labels.append("")

    opts = {}
    if autopct:
        cutoff = autopct if type(autopct) is not bool else 4
        opts.update(autopct=lambda p: '%1.1f%%' % p if p >= cutoff else "")
    else:
        opts.update(labeldistance=0.4)

    if others:
        plot_line = plot_line.rename({"Others": "Others (%1.2f%%)" % (others*100)})

    ax.pie(plot_line, labels=plotting_labels, radius=1.2,
           startangle=90, colors=color_list, **opts)
    ax.axis('equal')
    if legend:
        if nil_label is not None:
            ax.legend([nil_label if label == "<nil>" else label for label in plot_line.index.values], loc=(0.9, 0.2))
        else:
            ax.legend(plot_line.index.values, loc=(0.9, 0.2))


def plot_fix_point(table, ax, palette):
    """Plot a piechart representing the fixed point probability."""
    palette['no_fp'] = '#121212'
    prob_list = []
    color_list = []
    labels = []
    for i in range(len(table)):
        prob = table['Proba'][i]
        state = table['State'][i]
        prob_list.append(prob)
        color_list.append(persistent_color(palette, state))
        labels.append('FP '+str(i+1))
    prob_ns = 1 - sum(prob_list)
    if prob_ns > 0.01:
        prob_list.append(prob_ns)
        color_list.append(persistent_color(palette, "no_fp"))
        labels.append('no_fp')
    ax.pie(prob_list, labels=labels, colors=color_list)






from __future__ import print_function

from ..results.baseresult import BaseResult
from ..results.storedresult import StoredResult
import os
import tempfile
import subprocess
import sys
from random import random
import shutil
from sklearn.decomposition import PCA
from sklearn.manifold import TSNE
from sklearn.cluster import KMeans
import numpy as np
import multiprocessing
import pandas as pd
import matplotlib.pyplot as plt 
from re import match
import ast
import math
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import matplotlib.patches as mpatches

class EnsembleResult(BaseResult):
  
    # def __init__(self, models_files, cfg_filename, prefix="res", individual_results=False, random_sampling=False):
    def __init__(self, simulation, workdir=None, overwrite=False, prefix="res"):

        # self._cfg = cfg_filename
        self.workdir = workdir
        if workdir is None:
            self._path = tempfile.mkdtemp()
        else:
            self._path = workdir
            if os.path.exists(self._path) and overwrite:
                shutil.rmtree(self._path)
                os.mkdir(self._path)
        
            elif not os.path.exists(self._path):
                os.mkdir(self._path)


        self._cfg = os.path.join(self._path, "ensemble.cfg")

        BaseResult.__init__(self, self._path, simulation)
        self.prefix = prefix
        self.asymptotic_probtraj_distribution = None
        self.asymptotic_nodes_probtraj_distribution = None
        self._pcafig = None
        self._3dfig = None
        maboss_cmd = simulation.get_maboss_cmd()

        options = ["--ensemble"]
        if simulation.individual_results:
            options.append("--save-individual")

        if simulation.random_sampling:
            options.append("--random-sampling")

        cmd_line = [maboss_cmd] + options
                
        if len(simulation.individual_cfgs) > 0:
            os.mkdir(os.path.join(self._path, "models"))
            self.models_files = simulation.models_files
            cmd_line.append("--ensemble-istates")
            for model_file in self.models_files:
                cmd_line += ["-c", os.path.join(self._path, "models", os.path.basename(simulation.individual_cfgs[model_file]))]
                shutil.copyfile(model_file, os.path.join(self._path, "models", os.path.basename(model_file)))
                shutil.copyfile(simulation.individual_cfgs[model_file], os.path.join(self._path, "models", os.path.basename(simulation.individual_cfgs[model_file])))
        
        elif len(simulation.individual_istates) > 0:
            os.mkdir(os.path.join(self._path, "models"))
            cmd_line.append("--ensemble-istates")
            simulation.write_cfg(self._path, "ensemble.cfg")
            simulation.write_models(self._path)
            self.models_files = simulation.models_files

            for model_file in self.models_files:
                cmd_line += ["-c", os.path.join(self._path, "models", os.path.basename(simulation.individual_cfgs[os.path.basename(model_file)]))]
        
        else:
            simulation.write_cfg(self._path, "ensemble.cfg")
            simulation.write_models(self._path)
            self.models_files = simulation.models_files
            cmd_line += ["-c", self._cfg]     

        cmd_line += [
            "-o", self._path+'/'+self.prefix
        ] + self.models_files

        res = subprocess.Popen(
            cmd_line,
            stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )

        std_out, std_err = res.communicate()
        self._err = res.returncode
        if self._err != 0:
            print("Error, MaBoSS returned non 0 value", file=sys.stderr)
            print(std_err.decode())
        
        if len(std_out.decode()) > 0:
            print(std_out.decode())

    def get_thread_count(self):
        # TODO : Extracting it from the cfg
        return 6

    def get_fp_file(self):
        return os.path.join(self._path, "%s_fp.csv" % self.prefix)

    def get_probtraj_file(self):
        return os.path.join(self._path, "%s_probtraj.csv" % self.prefix)
        
    def get_statdist_file(self):
        return os.path.join(self._path, "%s_statdist.csv" % self.prefix)

    def load_individual_result(self, model):
        return StoredResult(self._path, self.prefix + "_model_" + str(model))

    def __del__(self):
        if self.workdir is None and os.path.exists(self._path):
            shutil.rmtree(self._path)

    def get_individual_states_probtraj(self, filter=None, cluster=None):
        """
        .. py:method:: Get a Panda Dataframe with the states final probability of each model
        
        :param filter: (optional) condition on the node distributions
        :param cluster: (optional) only get the result of a specified cluster, a list of ids
        
        """ 
        
        
        if self.asymptotic_probtraj_distribution is None:
            results = []
            for i, model in enumerate(self.models_files):
                results.append(self.load_individual_result(i))

            tables = []
            with multiprocessing.Pool(processes=self.get_thread_count()) as pool:
                tables = pool.starmap(get_single_individual_states_distribution, [(result, i) for i, result in enumerate(results)])
            self.asymptotic_probtraj_distribution = pd.concat(tables, axis=0, sort=False)
            self.asymptotic_probtraj_distribution.fillna(0, inplace=True)

        if filter is not None:
            return apply_filter(self.asymptotic_probtraj_distribution, filter, state=True)

        if cluster is not None:
            return self.asymptotic_probtraj_distribution.iloc[cluster, :]
        
        return self.asymptotic_probtraj_distribution

    def get_individual_nodes_probtraj(self, filter=None, cluster=None):
        """
        .. py:method:: Get a Panda Dataframe with the nodes final probability of each model
        
        :param filter: (optional) condition on the node distributions
        :param cluster: (optional) only get the result of a specified cluster, a list of ids
        
        """ 
        
        
        if self.asymptotic_nodes_probtraj_distribution is None:

            table = self.get_individual_states_probtraj()
            nodes = get_nodes(table.columns.values)
            with multiprocessing.Pool(processes=self.get_thread_count()) as pool:
                self.asymptotic_nodes_probtraj_distribution = pd.concat(
                    pool.starmap(get_single_individual_nodes_distribution, [(table, t_index, nodes) for t_index in table.index]), 
                    sort=False, axis=0
                )

        if filter is not None:
            return apply_filter(self.asymptotic_nodes_probtraj_distribution, filter)

        if cluster is not None:
            return self.asymptotic_nodes_probtraj_distribution.iloc[cluster, :]
        
        return self.asymptotic_nodes_probtraj_distribution

    def getByCondition(self, node_filter=None, state_filter=None):
        """
        .. py:method:: Filter the ensemble by condition on the node or state distribution
        
        :param node_filter: (optional) condition on the node distributions
        :param node_filter: (optional) condition on the state distributions
        
        """ 
        
        if node_filter is not None:
            indexes = self.get_individual_nodes_probtraj(node_filter).index.values
            labels = [0 if i not in indexes else 1 for i in range(len(self.models_files))]
            return indexes, labels
        elif state_filter is not None:
            indexes = self.get_individual_states_probtraj(state_filter).index.values
            labels = [0 if i not in indexes else 1 for i in range(len(self.models_files))]
            return indexes, labels
        

    def filterEnsembleByCondition(self, output_directory, node_filter=None, state_filter=None):
        """
        .. py:method:: Build an sub-ensemble from a condition on node or state distributions
        
        :param output_directory: directory in which to write the new ensemble
        :param node_filter: (optional) condition on the node distributions
        :param node_filter: (optional) condition on the state distributions
        
        """ 
         
        model_list = None
        if node_filter is not None:
            model_list = self.get_individual_nodes_probtraj(node_filter).index.values
              
        elif state_filter is not None:
            model_list = self.get_individual_states_probtraj(state_filter).index.values
              
        else:
            return None

        if not os.path.exists(output_directory):
            os.mkdir(output_directory)
        else:
            shutil.rmtree(output_directory)
            os.mkdir(output_directory)

        for model in model_list:
            shutil.copyfile(
                self.models_files[int(model)], 
                os.path.join(output_directory, os.path.basename(self.models_files[int(model)]))
            )
    
    def getKMeans(self, clusters=0):
        """
        .. py:method:: Perform a k-means clustering on the nodes distributions of each individual result 
        
        :param clusters: number of clusters
        
        :return: (dict associating cluster id to a list of models, labels of the clusters)
        
        """ 
        if clusters > 0:
            kmeans = KMeans(n_clusters=clusters).fit(self.get_individual_nodes_probtraj().values)
            indices = {}
            for i, label in enumerate(kmeans.labels_):
                if label in indices.keys():
                    array = indices[label]
                    array.append(i)
                    indices.update({label: array})
                else:
                    indices.update({label: [i]})

            return indices, kmeans.labels_

    def getStatesKMeans(self, clusters=0):
        """
        .. py:method:: Perform a k-means clustering on the state distributions of each individual result 
        
        :param clusters: number of clusters
        
        :return: (dict associating cluster id to a list of models, labels of the clusters)
        
        """ 
        if clusters > 0:
            kmeans = KMeans(n_clusters=clusters).fit(self.get_individual_states_probtraj().values)
            indices = {}
            for i, label in enumerate(kmeans.labels_):
                if label in indices.keys():
                    array = indices[label]
                    array.append(i)
                    indices.update({label: array})
                else:
                    indices.update({label: [i]})

            return indices, kmeans.labels_

    def filterEnsembleByCluster(self, output_directory, cluster):
        """
        .. py:method:: Build an sub-ensemble from a list of list of models
        
        :param output_directory: directory in which to write the new ensemble
        :param cluster: list of models to include in the new ensemble
        
        """ 
        if cluster is not None:
            if not os.path.exists(output_directory):
                os.mkdir(output_directory)
            else:
                shutil.rmtree(output_directory)
                os.mkdir(output_directory)

            for model in cluster:
                shutil.copyfile(
                    self.models_files[model], 
                    os.path.join(output_directory, os.path.basename(self.models_files[model]))
                )

    def plotStates3D(self, dims, figsize=(20, 12), compare=None, ax=None, **args):
        """
        .. py:method:: Plots the distribution of the ensemble individual results as a 3D object, for 3 given states.
        
        
        :param dims: list of the three states to plot
        :param figsize: (optional) tuple containing the size of the figure
        :param compare: (optional) other ensemble result for comparison
        :param ax: (optional) axes to plot on
        
        """
        if len(dims) == 3:
            table = self.get_individual_states_probtraj()
            
            if ax is None:
                self._3dfig = plt.figure(figsize=figsize)
                ax = self._3dfig.add_subplot(111, projection='3d')
            else:
                self._3dfig = ax.get_figure()
            
            if compare is not None:
                
                m_table = compare.get_individual_states_probtraj()
                
                # Here we need to make sure all tables have the same columns
                all_columns = set(list(table.columns) + list(m_table.columns) + dims)
                
                for column in all_columns:
                    if column not in table.columns.values:
                        table[column] = 0
                    if column not in m_table.columns.values:
                        m_table[column] = 0
                
                table = table[all_columns]
                m_table = m_table[all_columns]
                    
                values = table[dims].values
                m_values = m_table[dims].values
                ax.scatter(values[:, 0], values[:, 1], values[:, 2], **args)
                ax.scatter(m_values[:, 0], m_values[:, 1], m_values[:, 2], **args)
            else:
                values = table[dims].values
                ax.scatter(values[:, 0], values[:, 1], values[:, 2], **args)
                
            ax.set_xlabel(dims[0])
            ax.set_ylabel(dims[1])
            ax.set_zlabel(dims[2])

    def plotSteadyStatesDistribution(self, compare=None, labels=None, alpha=1, single_out=None, single_out_mutant=None, nil_label=None, compare_labels=None, **args):
        """
        .. py:method:: Plots the distribution of the ensemble individual results in PCA space
        
        :param compare: (optional) other ensemble simulation result, for comparison
        :param labels: (optional) list of colors to use for each model
        :param alpha: (optional) transparency of markers
        :param single_out: (optional) index of a model to highlight
        :param single_out_mutant: (optional) index of a model to highlight in the other ensemble simulation result
        :param nil_label: (optional) label for renaming the <nil> state
        :param compare_labels: (optional) labels to use in the legend
        
        """ 
        pca = PCA()
        table = self.get_individual_states_probtraj()
        if compare is not None:
            compare_table = compare.get_individual_states_probtraj()
            
            # Here we need to make sure all tables have the same columns
            all_columns = set(list(table.columns) + list(compare_table.columns))
            for column in all_columns:
                if column not in table.columns.values:
                    table[column] = 0
                if column not in compare_table.columns.values:
                    compare_table[column] = 0
            
            table = table[all_columns]
            compare_table = compare_table[all_columns]
    
            mat = table.values
            pca_res = pca.fit(mat)
            X_pca = pca.transform(mat)
            arrows_raw = (np.transpose(pca_res.components_[0:2, :]))
        
            c_pca = pca.transform(compare_table.values)
            
            self.plotPCA(
                pca, X_pca, 
                list(table.columns.values), list(table.index.values), labels, alpha,
                compare=c_pca,
                single_out=single_out, single_out_mutant=single_out_mutant, nil_label=nil_label, compare_labels=compare_labels,
                **args,
            )
        else:
            mat = table.values
            pca_res = pca.fit(mat)
            X_pca = pca.transform(mat)
            arrows_raw = (np.transpose(pca_res.components_[0:2, :]))
        
            self.plotPCA(
                pca, X_pca, 
                list(table.columns.values), list(table.index.values), labels, alpha, 
                single_out=single_out, nil_label=nil_label,
                **args
            )

    def plotSteadyStatesNodesDistribution(self, compare=None, labels=None, alpha=1, **args):
        """
        .. py:method:: Plots the nodes distribution of the ensemble individual results in PCA space
        
        :param compare: (optional) other ensemble simulation result, for comparison
        :param labels: (optional) list of colors to use for each model
        :param alpha: (optional) transparency of markers
        
        """ 
        
        pca = PCA()
        table = self.get_individual_nodes_probtraj()
        mat = table.values
        pca_res = pca.fit(mat)
        X_pca = pca.transform(mat)
        arrows_raw = (np.transpose(pca_res.components_[0:2, :]))

        if compare is not None:
            compare_table = compare.get_individual_nodes_probtraj()
            c_pca = pca.transform(compare_table.values)
            self.plotPCA(
                pca, X_pca,
                list(table.columns.values), list(table.index.values), labels, alpha,
                compare=c_pca, **args
            )
        else:
            self.plotPCA(
                pca, X_pca,
                list(table.columns.values), list(table.index.values), labels, alpha,
                **args
            )

    def plotPCA(self, 
        pca, X_pca, samples, features, colors=None, alpha=1, compare=None, single_out=None, single_out_mutant=None, nil_label=None, compare_labels=None,
        figsize=(20, 12), show_samples=False, show_features=True, ax=None, cutoff_arrows=None):
        
        if ax is None:
            self._pcafig = plt.figure(figsize=figsize)
            ax = self._pcafig.add_subplot(1,1,1)
        else:
            self._pcafig = ax.get_figure()
        
        patches = []
        if colors is None:
            ax.scatter(X_pca[:, 0], X_pca[:, 1], alpha=alpha, c='C0')
            patches.append(mpatches.Patch(color="C0", label=(compare_labels[0] if (compare_labels is not None and len(compare_labels) > 0) else None)))
            if single_out is not None:
                ax.scatter([X_pca[single_out, 0]], [X_pca[single_out, 1]], marker='o', facecolors='none', edgecolors='C0', s=200)
        else:
            legend = ["Cluster #%d" % (i + 1) for i in colors]

            c_colors = ["C%d" % color for color in colors]
            
            scatter = ax.scatter(X_pca[:, 0], X_pca[:, 1], c=c_colors, s=50, alpha=alpha, label=colors)

            # build the legend

            patches = []
            for cluster in set(colors):
                patches.append(
                    mpatches.Patch(color="C%d" % list(set(colors))[cluster], label='Cluster #%d' % (cluster + 1))
                )

            legend = plt.legend(handles=patches)

        if compare is not None:
            ax.scatter(compare[:, 0], compare[:, 1], alpha=alpha, c='C1')
            patches.append(mpatches.Patch(color="C1", label=(compare_labels[1] if (compare_labels is not None and len(compare_labels) > 1) else None)))
            if single_out_mutant is not None:
                ax.scatter([compare[single_out_mutant, 0]], [compare[single_out_mutant, 1]], marker='o', facecolors='none', edgecolors='C1', s=200)
            
        ax.set_xlabel("PC{} ({}%)".format(1, round(pca.explained_variance_ratio_[0] * 100, 2)))
        ax.set_ylabel("PC{} ({}%)".format(2, round(pca.explained_variance_ratio_[1] * 100, 2)))
        legend = plt.legend(handles=patches)

        arrows_raw = pca.components_[0:2, :].T
        
        max_x_arrows = max(arrows_raw[:, 0])
        min_x_arrows = min(arrows_raw[:, 0])

        if compare is None:
            max_x_values = max(X_pca[:, 0])
            min_x_values = min(X_pca[:, 0])
        else:
            max_x_values = max(max(X_pca[:, 0]), max(compare[:, 0]))
            min_x_values = min(min(X_pca[:, 0]), min(compare[:, 0]))
        
        max_y_arrows = max(arrows_raw[:, 1])
        min_y_arrows = min(arrows_raw[:, 1])

        if compare is None:
            max_y_values = max(X_pca[:, 1])
            min_y_values = min(X_pca[:, 1])
        else:
            max_y_values = max(max(X_pca[:, 1]), max(compare[:, 1]))
            min_y_values = min(min(X_pca[:, 1]), min(compare[:, 1]))
  
        if show_samples:
            for i, txt in enumerate(features):
                ax.annotate(txt, (X_pca[i, 0], X_pca[i, 1]))
  
        if show_features:
            for i, v in enumerate(arrows_raw):
                if cutoff_arrows is None or math.sqrt(math.pow(v[0], 2) + math.pow(v[1], 2)) > cutoff_arrows:
                    ax.arrow(0, 0, v[0], v[1], width=0.003, color='red')
                    if samples[i] == "<nil>" and nil_label is not None:
                        ax.text(v[0], v[1], nil_label, color='black', ha='right', va='top')
                    else:
                        ax.text(v[0], v[1], samples[i], color='black', ha='right', va='top')

            ax.set_xlim(min(min_x_values, min_x_arrows)*1.2, max(max_x_values, max_x_arrows)*1.2)
            ax.set_ylim(min(min_y_values, min_y_arrows)*1.2, max(max_y_values, max_y_arrows)*1.2)
        else:
            ax.set_xlim(min_x_values*1.2, max_x_values*1.2)
            ax.set_ylim(min_y_values*1.2, max_y_values*1.2)

    def plotTSNESteadyStatesNodesDistribution(self, node_filter=None, state_filter=None, clusters={}, perplexity=50, n_iter=2000, **args):
        """
        .. py:method:: Plots the nodes distribution of the ensemble individual results in T-SNE space
        
        :param node_filter: (optional) filter in node distribution to highlight a sub-ensemble of models
        :param node_filter: (optional) filter in state distribution to highlight a sub-ensemble of models
        :param cluster: (optional) dict with, for each model, the id of the cluster if belongs to
        :param perplexity: (optional) hyper-parameter of T-SNE (default=50)
        :param n_iter: (optional) default parameter of T-SNE (default=2000)
        
        """ 
        
        pca = PCA()
        table = self.get_individual_nodes_probtraj()
        
        model = TSNE(perplexity=perplexity, n_iter=n_iter, n_iter_without_progress=n_iter*0.5)   
        res = model.fit_transform(table.values)

        if node_filter is None and state_filter is None:
            if len(clusters) == 0:
                fig = plt.figure(**args)
                plt.scatter(res[:, 0], res[:, 1])
            else:
                fig = plt.figure(**args)
                for i, cluster in clusters.items():
                    plt.scatter(res[cluster, 0], res[cluster, 1], color="C%d" % i)

        else:
            fig = plt.figure(**args)
            filtered, _ = self.getByCondition(node_filter=node_filter, state_filter=state_filter)
            not_filtered = list(set(range(len(self.models_files))).difference(set(filtered)))
            
            plt.scatter(res[filtered, 0], res[filtered, 1], color='r')
            plt.scatter(res[not_filtered, 0], res[not_filtered, 1], color='b')

    def plotTSNESteadyStatesDistribution(self, node_filter=None, state_filter=None, clusters={}, perplexity=50, n_iter=2000, **args):
        """
        .. py:method:: Plots the states distribution of the ensemble individual results in T-SNE space
        
        :param node_filter: (optional) filter in node distribution to highlight a sub-ensemble of models
        :param node_filter: (optional) filter in state distribution to highlight a sub-ensemble of models
        :param cluster: (optional) dict with, for each model, the id of the cluster if belongs to
        :param perplexity: (optional) hyper-parameter of T-SNE (default=50)
        :param n_iter: (optional) default parameter of T-SNE (default=2000)
        
        """ 
        pca = PCA()
        table = self.get_individual_states_probtraj()
        
        model = TSNE(perplexity=perplexity, n_iter=n_iter, n_iter_without_progress=n_iter*0.5)   
        res = model.fit_transform(table.values)



        if node_filter is None or state_filter is None:
            if len(clusters) == 0:
                fig = plt.figure(**args)
                plt.scatter(res[:, 0], res[:, 1])
            else:
                fig = plt.figure(**args)
                for i, cluster in clusters.items():
                    plt.scatter(res[cluster, 0], res[cluster, 1], color="C%d" % i)

        else:
            fig = plt.figure(**args)
            filtered, _ = self.getByCondition(node_filter=node_filter, state_filter=state_filter)
            not_filtered = list(set(range(len(self.models_files))).difference(set(filtered)))
            
            plt.scatter(res[filtered, 0], res[filtered, 1], color='r')
            plt.scatter(res[not_filtered, 0], res[not_filtered, 1], color='b')


def fix_order(string):
    return " -- ".join(sorted(string.split(" -- ")))

def get_single_individual_states_distribution(result, i):
    if os.path.getsize(result.get_probtraj_file()) > 0:
        raw_table_states = result.get_last_states_probtraj()
        table_states = result.get_last_states_probtraj()
        table_states.rename(index={table_states.index[0]: i}, inplace=True)
        rename_columns = {col: fix_order(col) for col in table_states.columns}
        table_states.rename(columns=rename_columns, inplace=True)
        return table_states

def get_nodes(states):
    nodes = set()
    for s in states:
        if s != '<nil>':
            nds = s.split(' -- ')
            for nd in nds:
                nodes.add(nd)
    return nodes

def get_single_individual_nodes_distribution(table, index, nodes):
    ntable = pd.DataFrame(np.zeros((1, len(nodes))), index=[index], columns=nodes)
    for i, row in enumerate(table):
        state = table.columns[i]
        if state != "<nil>":
            t_nodes = state.split(" -- ")
            for node in t_nodes:
                ntable.loc[index, node] += table.loc[index, state]
                
    return ntable

def apply_filter(data, filter, state=False):

    if state: 
        filter = filter.replace(" -- ", "")
        dict_states = {column.replace(" -- ", ""): column for column in list(data.columns)}
        formula = ast.parse(filter)
        return parse_ast(formula.body[0].value, data, dict_states)
    else:
        formula = ast.parse(filter)
        return parse_ast(formula.body[0].value, data)

def parse_ast(t_ast, data, ds=None):
    if isinstance(t_ast, ast.BoolOp):
        
        if isinstance(t_ast.op, ast.And):
            
            values = [parse_ast(tt_ast, data, ds) for tt_ast in t_ast.values]
            t_data = pd.merge(values[0], values[1], how="inner", on=list(values[0].columns), left_index=True, right_index=True)

            for i in range(2, len(values)):
                t_data = pd.merge(t_data, values[i], how="inner", on=list(values[0].columns), left_index=True, right_index=True)

            return t_data
        elif isinstance(t_ast.op, ast.Or):
            
            values = [parse_ast(tt_ast, data, ds) for tt_ast in t_ast.values]
            t_data = pd.merge(values[0], values[1], how="outer", on=list(values[0].columns), left_index=True, right_index=True)

            for i in range(2, len(values)):
                t_data = pd.merge(t_data, values[i], how="outer", on=list(values[0].columns), left_index=True, right_index=True)

            return t_data
        
    elif isinstance(t_ast, ast.Compare):

        res_dict = {
            ast.Lt : data[parse_ast(t_ast.left, data, ds) < parse_ast(t_ast.comparators[0], data, ds)],
            ast.Gt : data[parse_ast(t_ast.left, data, ds) > parse_ast(t_ast.comparators[0], data, ds)],
            ast.LtE : data[parse_ast(t_ast.left, data, ds) <= parse_ast(t_ast.comparators[0], data, ds)],
            ast.GtE : data[parse_ast(t_ast.left, data, ds) >= parse_ast(t_ast.comparators[0], data, ds)],
            ast.Eq : data[parse_ast(t_ast.left, data, ds) == parse_ast(t_ast.comparators[0], data, ds)],
            ast.NotEq : data[parse_ast(t_ast.left, data, ds) != parse_ast(t_ast.comparators[0], data, ds)],
        }
        return res_dict[type(t_ast.ops[0])]

    elif isinstance(t_ast, ast.Num):
        return t_ast.n

    elif isinstance(t_ast, ast.Name):
        if ds is not None:
            if t_ast.id in ds.keys():
                return data[ds[t_ast.id]]
            else:
                return pd.Series(np.zeros(data.iloc[:, 0].shape), index=data.index)      
        else:
            return data[t_ast.id]
    
        
        
            



menu = [
    {'name': 'Load MaBoSS file',
     'snippet': ["masim = maboss.load(\"filename.bnd\", \"filename.cfg\")"]},

    {'name': 'Network',
     'sub-menu': [
         {'name': 'Set istate',
          'snippet': ["maboss.wg_set_istate(masim)"]},
         {'name': 'Get initial state',
          'snippet': ["masim.get_initial_state()"]}
     ]},

    {'name': 'Simulation',
     'sub-menu': [
          {'name': 'Create mutant',
           'snippet': ["maboss.wg_make_mutant(masim)"]},

          {'name': 'Run',
           'snippet': ["simres = masim.run()"]},

          {'name': 'Set initial states',
           'snippet': ["masim.network.set_istate([\"mygenelist\"],{})"]},

          {'name': 'Set output',
           'snippet': ["maboss.wg_set_output(masim)"]},

     ]},

    {'name': 'Results',
     'sub-menu': [
                  {'name': 'Save results',
                   'snippet': ["simres.save(\"filename\")"]},

                  {'name': 'Plot piechart',
                   'snippet': ["simres.plot_piechart()"]},

                  {'name': 'Plot trajectory',
                   'snippet': ["simres.plot_trajectory()"]}]
     },
     {'name': 'Conversion',
     'sub-menu': [
                  {'name': 'Convert to biolqm',
                   'snippet': ["biolqm_model = maboss.to_biolqm(masim)"]},

                  {'name': 'Convert to minibn',
                   'snippet': ["minibn_model = maboss.to_minibn(masim)"]},
     ]},
     "---",
     {"name": "Documentation",
        "external-link": "http://pymaboss.readthedocs.io"}
]

toolbar = None
js_api = {}
